' *!* make sure you build this example with the "Build GUI App" option turned off

Strict
Framework BRL.Blitz
Import BRL.StandardIO
Import gg.IrrBMAX

Rem
In this tutorial, I will show how To collision detection with the Irrlicht Engine. 
I will describe 3 methods: Automatic collision detection For moving through 3d worlds
with stair climbing And sliding, manual triangle picking And manual
scene node picking.

To start, we take the program from tutorial 2, which loaded And displayed a quake 3
level. We will use the level To walk in it And To pick triangles from it. In addition
we'll place 3 animated models into it for scene node picking. The following code 
starts up the engine And loads
a quake 3 level. I will Not explain it, because it should already be known from tutorial
2.
EndRem

' let user Select driver Type

Local driverType:Int=EDT_NULL
Local strInput:String=""

strInput$= ..
	Input("Please select the driver you want for this example:"+Chr(13)+Chr(10)+ ..
	" (a) Direct3D 9.0c"+Chr(13)+Chr(10)+ ..
	" (b) Direct3D 8.1"+Chr(13)+Chr(10)+ ..
	" (c) OpenGL 1.2"+Chr(13)+Chr(10)+ ..
	" (d) Software Renderer"+Chr(13)+Chr(10)+ ..
	" (e) Apfelbaum Software Renderer"+Chr(13)+Chr(10)+ ..
	" (f) NullDevice"+Chr(13)+Chr(10)+ ..
	" (otherKey) Exit"+Chr(13)+Chr(10)+Chr(13)+Chr(10))

Select strInput$
	Case "a"
		driverType=EDT_DIRECT3D9
	Case "b"
		driverType=EDT_DIRECT3D8
	Case "c"
		driverType=EDT_OPENGL
	Case "d"
		driverType=EDT_SOFTWARE
	Case "e"
		driverType=EDT_SOFTWARE2
	Case "f"
		driverType=EDT_NULL
	Default
		Return 
EndSelect

' create device

Local device:T_irrIrrlichtDevice= ..
	T_irrIrrlichtDevice.create(driverType,T_irrDimension2d_s32.create(640,480),16,False)

If device.handle=0 Then Return ' could Not create selected driver.

Local driver:T_irrIVideoDriver=device.getVideoDriver()
Local smgr:T_irrISceneManager=device.getSceneManager()
	
device.getFileSystem().addZipFileArchive("../media/map-20kdm2.pk3")
Local q3levelmesh:T_irrIAnimatedMesh=smgr.getMesh("20kdm2.bsp")	
Local q3node:T_irrISceneNode
	
If (q3levelmesh.isValid()) Then q3node=smgr.addOctTreeSceneNode(q3levelmesh.getMesh(0))

Rem
So far so good, we've loaded the quake 3 level like in tutorial 2. Now, here
comes something different: We create a triangle selector. A triangle selector
is a class which can fetch the triangles from scene nodes For doing different
things with them, For example collision detection. There are different triangle
selectors, And all can be created with the ISceneManager. In this example,
we create an OctTreeTriangleSelector, which optimizes the triangle output a l
little bit by reducing it like an octree. This is very useful For huge meshes
like quake 3 levels.
Afte we created the triangle selector, we attach it To the q3node. This is Not
necessary, but in this way, we do Not need To care For the selector, For example
dropping it after we do Not need it anymore.
EndRem

Local selector:T_irrITriangleSelector

If (q3node.isValid())
	q3node.setPosition(T_irrVector3df.createFromVals(-1370,-130,-1400))
	
	selector = smgr.createOctTreeTriangleSelector(q3levelmesh.getMesh(0), q3node, 128)
	q3node.setTriangleSelector(selector)
	' BMX types automatically drop if needed
	'selector.drop()
EndIf

Rem
We add a first person shooter camera To the scene For being able To move in the quake 3
level like in tutorial 2. But this, time, we add a special animator To the 
camera: A Collision Response animator. This thing modifies the scene node To which
it is attached To in that way, that it may no more move through walls And is affected
by gravity. The only thing we have To tell the animator is how the world looks like,
how big the scene node is, how gravity And so on. After the collision response animator
is attached To the camera, we do Not have To do anything more For collision detection,
anything is done automaticly, all other collision detection code below is For picking.
And please note another cool feature: The collsion response animator can be attached
also To all other scene nodes, Not only To cameras. And it can be mixed with other
scene node animators. In this way, collision detection And response in the Irrlicht
engine is really, really easy.
Now we'll take a closer look on the parameters of createCollisionResponseAnimator().
The first parameter is the TriangleSelector, which specifies how the world, against
collision detection is done looks like. The second parameter is the scene node, which
is the Object, which is affected by collision detection, in our Case it is the camera.
The third defines how big the Object is, it is the radius of an ellipsoid. Try it out 
And change the radius To smaller values, the camera will be able To move closer To walls
after this. The Next parameter is the direction And speed of gravity. You could
set it To (0,0,0) To disable gravity. And the last value is just a translation: Without
this, the ellipsoid with which collision detection is done would be around the camera,
And the camera would be in the middle of the ellipsoid. But as human beings, we are 
used To have our eyes on top of the body, with which we collide with our world, Not
in the middle of it. So we place the scene node 50 units over the center of the 
ellipsoid with this parameter. And that's it, collision detection works now. 
EndRem

Local camera:T_irrICameraSceneNode=smgr.addCameraSceneNodeFPS(Null,100.0,300.0,-1,Null,True)
camera.setPosition(T_irrVector3df.createFromVals(-100,50,-150))

Local anim:T_irrISceneNodeAnimator=smgr.createCollisionResponseAnimator( ..
		selector, camera, T_irrVector3df.createFromVals(30,50,30), ..
		T_irrVector3df.createFromVals(0,-3,0), ..
		T_irrVector3df.createFromVals(0,50,0))
		
camera.addAnimator(anim)
' BMX types automatically drop if needed
'anim.drop()

Rem
Because collision detection is no big deal in irrlicht, I'll describe how to
do two different types of picking in the Next section. But before this,
I'll prepare the scene a little. I need three animated characters which we 
could pick later, a dynamic light For lighting them,
a billboard For drawing where we found an intersection,	And, yes, I need To
get rid of this mouse cursor. :)
EndRem

' disable mouse cursor
device.getCursorControl().setVisible(False)

' add billboard
Local bill:T_irrIBillboardSceneNode=smgr.addBillboardSceneNode()
bill.setMaterialType(EMT_TRANSPARENT_ADD_COLOR)
bill.setMaterialTexture(0,driver.getTexture("../media/particle.bmp"))
bill.setMaterialFlag(EMF_LIGHTING, False)
bill.setMaterialFlag(EMF_ZBUFFER, False)
bill.setSize(T_irrDimension2d_f32.create(20.0, 20.0))

' add 3 animated faeries.
Local material:T_irrSMaterial=T_irrSMaterial.create()
material.setTexture1(driver.getTexture("../media/faerie2.bmp"))
material.setLighting(True)

Local node:T_irrIAnimatedMeshSceneNode
Local faerie:T_irrIAnimatedMesh=smgr.getMesh("../media/faerie.md2")

If (faerie.isValid())
	node=smgr.addAnimatedMeshSceneNode(faerie)
	node.setPosition(T_irrVector3df.createFromVals(-70,0,-90))
	node.setMD2AnimationFromType(EMAT_RUN)
	node.setMaterial(0,material)

	node=smgr.addAnimatedMeshSceneNode(faerie)
	node.setPosition(T_irrVector3df.createFromVals(-70,0,-30))
	node.setMD2AnimationFromType(EMAT_SALUTE)
	node.setMaterial(0,material)

	node=smgr.addAnimatedMeshSceneNode(faerie)
	node.setPosition(T_irrVector3df.createFromVals(-70,0,-60))
	node.setMD2AnimationFromType(EMAT_JUMP)
	node.setMaterial(0,material)
EndIf

material.setTexture1()
material.setLighting(False)

' Add a light
smgr.addLightSceneNode(Null, T_irrVector3df.createFromVals(-60,100,400), ..
	T_irrSColorf.createFromRGBA(1.0,1.0,1.0,1.0),600.0)

Rem
For Not making it To complicated, I'm doing picking inside the drawing loop.
We take two pointers For storing the current And the last selected scene node And 
start the loop.
EndRem

Local selectedSceneNode:T_irrISceneNode
Local lastSelectedSceneNode:T_irrISceneNode=Null
	
Local lastFPS:Int = -1
Local fps:Int = 0
Local str:String=""

Local line:T_irrLine3d_f32=T_irrLine3d_f32.create()
Local intersection:T_irrVector3df=T_irrVector3df.create()
Local tri:T_irrTriangle3df=T_irrTriangle3df.create()

While(device.run())
	If (device.isWindowActive())

		driver.beginScene(True, True)
		smgr.drawAll()

		Rem
		After we've drawn the whole scene whit smgr->drawAll(), we'll do the first
		picking: We want To know which triangle of the world we are looking at. In addition,
		we want the exact point of the quake 3 level we are looking at.
		For this, we create a 3d line starting at the position of the camera And going 
		through the lookAt-target of it. Then we ask the collision manager If this line
		collides with a triangle of the world stored in the triangle selector. If yes,
		we draw the 3d triangle And set the position of the billboard To the intersection 
		point.
		EndRem

		line.setLineFromVects(camera.getPosition(), ..
			line.getStart().Plus((camera.getTarget().Minus(line.getStart())).normalize().Mult(1000.0)))

		If (smgr.getSceneCollisionManager().getCollisionPoint( ..
			line, selector, intersection, tri))

			bill.setPosition(intersection)
				
			driver.SetTransform(ETS_WORLD,T_irrMatrix4.create())
			driver.setMaterial(material)
			driver.draw3DTriangle(tri,T_irrSColor.createFromVals(0,255,0,0))
		EndIf

		Rem
		Another Type of picking supported by the Irrlicht Engine is scene node picking
		based on bouding boxes. Every scene node has got a bounding box, And because of
		that, it's very fast for example to get the scene node which the camera looks
		at. Again, we ask the collision manager For this, And If we've got a scene node,
		we highlight it by disabling Lighting in its material, If it is Not the 
		billboard Or the quake 3 level.
		EndRem
		
		selectedSceneNode=smgr.getSceneCollisionManager().getSceneNodeFromCameraBB(camera)

		If (lastSelectedSceneNode<>Null And lastSelectedSceneNode.isValid())
			lastSelectedSceneNode.setMaterialFlag(EMF_LIGHTING, True)
		EndIf
		
		If (selectedSceneNode.Compare(q3node) Or selectedSceneNode.Compare(bill))
			selectedSceneNode = Null
		EndIf
		
		If (selectedSceneNode)
			selectedSceneNode.setMaterialFlag(EMF_LIGHTING, False)
		EndIf
		lastSelectedSceneNode = selectedSceneNode

		Local triselector:T_irrITriangleSelector=smgr.createTriangleSelector(faerie.getMesh(0),node)


		Rem
		That's it, we just have to finish drawing.
		EndRem

		driver.endScene()

		fps = driver.getFPS()

		If (lastFPS <> fps) Then
			str$ = "Collision detection example - Irrlicht Engine ["
			str$ :+ driver.getName()
			str$ :+ "] FPS:"
			str$ :+ fps

			device.setWindowCaption(str$)
			lastFPS = fps
		EndIf
	EndIf
Wend

device.drop()
